package net.sf.corn.httpclient;

import java.io.BufferedReader;
import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookiePolicy;
import java.net.HttpCookie;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.URI;
import java.net.URL;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;
import java.util.zip.ZipInputStream;

/**
 * This class is a simple client for http requests. It's pretty generic one 
 * to make any Http request. It uses {@link TSAuthenticator} to handle security 
 * credentials, thus http requests that requires authentication can be done.
 * Cookies and http request header parameters are also supported.
 * <p><code>
 * HttpClient client = new HttpClient(new URI("https://www.googleapis.com/books/v1/volumes"); </br>
 * //optional proxy creds</br>
 * client.setProxyCredentials(yourProxyUser,yourProxyPassword);</br>
 * //optional request creds</br>
 * client.setProxyCredentials(yourUser,yourPassword);</br>
 * HttpResponse response = client.sendData(HTTP_METHOD.GET,"q=quilting");</br>
 * if(response.hasError()){</br>
 * &nbsp System.out.println("Error:"+response.getErrorMessage()+","+response.getErrorData());</br>
 * }else{</br>
 * &nbsp System.out.println("Response Code:"+response.getResponseCode());</br>
 * &nbsp System.out.println("Response Data:"+response.getResponseData());</br>
 * &nbsp System.out.println("Cookies:"+response.getCookies());</br>
 * &nbsp System.out.println("Headers:"+response.getHeaderFields());</br>
 * }</br>
 * </code></p>
 * @author Serhat Dirik
 * @see TSAuthenticator
 */
public class HttpClient {
	// Inner Types
	public static enum HTTP_METHOD {
		GET, POST, DELETE, OPTIONS, HEAD, PUT, TRACE, CONNECT;
	}
	
	// Constants
	public  static final String USER_AGENT_CHROME_28_0_1468_0 = "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1468.0 Safari/537.36";
	public  static final String USER_AGENT_FIREFOX_23_0 = "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:23.0) Gecko/20130406 Firefox/23.0";
	public  static final String USER_AGENT_IE_10_6 = "Mozilla/5.0 (compatible; MSIE 10.6; Windows NT 6.1; Trident/5.0; InfoPath.2; SLCC1; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729; .NET CLR 2.0.50727) 3gpp-gba UNTRUSTED/1.0";
	public  static final String USER_AGENT_ANDROID_WEBKIT = "Mozilla/5.0 (Linux; U; Android 4.0.3; ko-kr; LG-L160L Build/IML74K) AppleWebkit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30";
	
	private static final String UTF8_CHARSET = "UTF-8";
	private static final String PROP_ACCEPT_ENCODING = "Accept-Encoding";
	private static final String PROP_USER_AGENT = "User-Agent";
	private static final String PROP_ACCEPT = "Accept";
	private static final String PROP_ACCEPT_LANGUAGE = "Accept-Language";
	private static final String PROP_ACCEPT_CHARSET = "Accept-Charset";
	private static final String PROP_CONTENT_TYPE = "Content-Type";
	private static final String PROP_KEEP_ALIVE = "Keep-Alive";
	private static final String PROP_CONNECTION = "Connection";
	private static final String PROP_VAL_CONNECTION_KEEP_ALIVE = "keep-alive";
	private static final String PROP_VAL_CONNECTION_CLOSE = "close";
	private static final String ENCODING_GZIP = "gzip";
	private static final String ENCODING_DEFLATE = "deflate";
	private static final String ENCODING_ZIP = "zip";

	// Variables
	private String charset = UTF8_CHARSET;// +";q=0.7,*;q=0.7";
	private Map<String, String> additionalRequestProperty = new HashMap<String, String>();
	private SimpleCookieStore cookieStore = new SimpleCookieStore();
	private boolean followRedirects = true;
	private int timeout = -1;
	private URI uri;
	private boolean keepAlive = false;
	private String uname;
	private String password;
	private Proxy.Type proxyType = null;
	private String proxyHostName = null;
	private Integer proxyPort = null;
	private String proxyUname;
	private String proxyPassword;
	private String acceptedEncodings = "gzip,deflate";
	private String user_agent = "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.0.3) Gecko/20060426 Firefox/1.5.0.3";
	private String acceptedType = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8";
	private String acceptedLanguage = " en-us,en;q=0.5";
	private String contentType = "application/x-www-form-urlencoded";
	private Integer keepAliveTime = 300;
	private int bufferSize = 4096;
	private boolean cacheEnabled=false;
	
	/**
	 * Default constructor that creates a client for the given URI
	 */
	public HttpClient(URI uri) {
		if (uri == null) {
			throw new IllegalArgumentException("uri is null!");
		}
		this.uri = uri;
		// Set Authenticator
		TSAuthenticator.setAsDefault();
		// Set cookie manager
		CookieHandler.setDefault(new CookieManager(cookieStore,
				CookiePolicy.ACCEPT_ALL));
	}

	/**
	 * Gives the URI target of this client
	 * @return URI
	 */
	public URI getURI() {
		return uri;
	}

	/**
	 * Gives the charset 
	 */
	public String getCharset() {
		return charset;
	}

	/**
	 * Sets the charset of the client
	 * @param charset Characterset
	 * @throws UnsupportedEncodingException
	 */
	public void setCharset(String charset) throws UnsupportedEncodingException {
		URLEncoder.encode("test", charset);
		this.charset = charset;
	}

	/**
	 * Puts the given name,value pair to the http header
	 * 
	 */
	public void putAdditionalRequestProperty(String name, String value) {
		additionalRequestProperty.put(name, value);
	}

	/**
	 * Clear all additional header fields
	 */
	public void clearAdditionalRequestProperties() {
		additionalRequestProperty.clear();
	}

	/**
	 * Sets the read time out value
	 */
	public void setReadTimeout(int miliseconds) {
		timeout = miliseconds;
	}

	/**
	 * Read Time out value of the client
	 * @return time out in miliseconds
	 */
	public int getReadTimeout() {
		return timeout;
	}

	public void setProxySetting(Proxy.Type type, String hostName,int port){
		this.proxyType = type;
		this.proxyHostName = hostName;
		this.proxyPort = port;
	}
	/**
	 * Sets the proxy credentials
	 * @param uname user name
	 * @param password password
	 */
	public void setProxyCredentials(String uname, String password) {
		proxyUname = uname;
		proxyPassword = password;
	}
	
	/**
	 * Set security credentials of the target URI. Both basic and digest methods are supported
	 * @param uname
	 * @param password
	 */
	public void setCredentials(String uname, String password) {
		this.uname = uname;
		this.password = password;
	}

	/**
	 * Tells whether keep alive parameter is set or not
	 * @return true if set
	 */
	protected boolean isKeepAlive() {
		return keepAlive;
	}

	/**
	 * on/off keep alive parameter
	 * @param keepAlive true/false
	 */
	public void setKeepAlive(boolean keepAlive) {
		this.keepAlive = keepAlive;
	}

	/**
	 * Clears all cookie values
	 */
	public void clearCookies() {
		cookieStore.removeAll();
	}

	/**
	 * gives all cookie values
	 * @return List of cookies
	 */
	public List<HttpCookie> getCookies() {
		return cookieStore.get(uri);
	}

	/**
	 * gives all cookie values
	 * @return cookies as one string
	 */
	public String getCookiesString() {
		return cookieStore.getCookiesString(uri);
	}

	/**
	 * Gives the requested cookie value
	 * 
	 * @param name    name of the cookie
	 * @return cookie value
	 */
	public HttpCookie getCookie(String name) {
		List<HttpCookie> cookies = getCookies();
		if (cookies != null) {
			for (HttpCookie cookie : cookies) {
				if (cookie.getName().equals(name))
					return cookie;
			}
		}
		return null;
	}

	/**
	 * Adds all cookies in the given String to the client
	 * @param cookies for example 'ppkcookie1=testcookie; expires=Thu, 2 Aug 2001 20:47:11 UTC; path=/'
	 */
	public void addCookies(String cookies) {
		cookieStore.addCookiesString(uri, cookies);
	}
	
	/**
	 * Adds a cookie to the client
	 * @param cookie an http cookie
	 */
	public void addCookie(HttpCookie cookie) {
		// if(name.equals("domain") ||name.equals("path") ||
		// name.equals("expires")) return;
		cookieStore.add(uri, cookie);
	}

	/**
	 * Adds a cookies to the client
	 * @param cookies list of cookies
	 */
	public void addCookies(List<HttpCookie> cookies) {
		// if(name.equals("domain") ||name.equals("path") ||
		// name.equals("expires")) return;
		for (HttpCookie httpCookie : cookies) {
			cookieStore.add(uri,httpCookie);	
		}
	}
	
	/**
	 * Removes the given cookie from the client
	 * @param name cookie name
	 */
	public void removeCookie(String name) {
		HttpCookie cookie = getCookie(name);
		if (cookie != null)
			cookieStore.remove(uri, cookie);
	}
	
	/**
	 * Reset all previous cookie values and puts the given cookies string 
	 * @param cookies for example 'ppkcookie1=testcookie; expires=Thu, 2 Aug 2001 20:47:11 UTC; path=/'
	 */
	public void setCookies(String cookies) {
		cookieStore.removeAll();
		cookieStore.addCookiesString(uri, cookies);
	}
	
	/**
	 * Tells whether this client will fallow HTTP redirect actions or not
	 * @return
	 */
	public boolean isFollowRedirects() {
		return followRedirects;
	}

	/**
	 * Sets if the call should follow HTTP 302 redirects
	 * @param followRedirects true if you want to enable fallows
	 */
	public void setFollowRedirects(boolean followRedirects) {
		this.followRedirects = followRedirects;
	}

	/**
	 * Gives all additional request parameters to be added to HTTP Header
	 * @return key,value pairs of additional header properties
	 */
	public Map<String, String> getAdditionalRequestProperties() {
		return additionalRequestProperty;
	}
	
	/**
	 * @return Accepted Encoding parameter
	 */
	public String getAcceptedEncodings() {
		return acceptedEncodings;
	}
	/**
	 * @return User Agent String
	 */
	public String getUserAgent() {
		return user_agent;
	}
	/**
	 * @return Accepted Types String
	 */
	public String getAcceptedType() {
		return acceptedType;
	}
	/**
	 * @return Accepted Language Parameter
	 */
	public String getAcceptedLanguage() {
		return acceptedLanguage;
	}
	/**
	 * @return Expected content type
	 */
	public String getContentType() {
		return contentType;
	}
	/**
	 * @return keep alive time parameter which useful when keep alive enabled
	 */
	public Integer getKeepAliveTime() {
		return keepAliveTime;
	}
	/**
	 * @return Buffer size to read output
	 */
	public int getBufferSize() {
		return bufferSize;
	}
	/**
	 * Sets additional header parameters 
	 * @param additionalRequestProperty header parameters as Map
	 */
	public void setAdditionalRequestProperty(
			Map<String, String> additionalRequestProperty) {
		this.additionalRequestProperty = additionalRequestProperty;
	}
	/**
	 * Sets the accepted encodings 
	 */
	public void setAcceptedEncodings(String acceptedEncodings) {
		this.acceptedEncodings = acceptedEncodings;
	}
	/**
	 * Sets the user agent to simulate a browser
	 * @param user_agent sample agent string are given as constants
	 */
	public void setUserAgent(String user_agent) {
		this.user_agent = user_agent;
	}

	/**
	 * Sets the accepted types client parameter 
	 */
	public void setAcceptedType(String acceptedType) {
		this.acceptedType = acceptedType;
	}
	/**
	 * Sets the acceptable human languages for response
	 * @param acceptedLanguage
	 */
	public void setAcceptedLanguage(String acceptedLanguage) {
		this.acceptedLanguage = acceptedLanguage;
	}
	/**
	 * The MIME type of the body of the request 
	 * @param contentType
	 */
	public void setContentType(String contentType) {
		this.contentType = contentType;
	}
	
	/**
	 * Although this parameter not is the HTTP spec, its useful to tell the expected timeout to the server 
	 * @param keepAliveTime
	 */
	public void setKeepAliveTime(Integer keepAliveTime) {
		this.keepAliveTime = keepAliveTime;
	}

	/**
	 * Sets the buffer size to read outputs 
	 * @param bufferSize a pozitive integer greater or equal to 4096
	 */
	public void setBufferSize(int bufferSize) {
		this.bufferSize = bufferSize < 4096 ? 4096 : bufferSize;
	}

	/**
	 * Tells whether the client cache is enabled or not
	 * @return true if the cache is enabled
	 */
	public boolean isCacheEnabled() {
		return cacheEnabled;
	}

	/**
	 * Sets the client cache enabled or disabled
	 * @param cacheEnabled true if the cache would be enabled
	 */
	public void setCacheEnabled(boolean cacheEnabled) {
		this.cacheEnabled = cacheEnabled;
	}

	/**
	 * Gives a connection to requested url
	 */
	@SuppressWarnings("restriction")
	private HttpURLConnection getConnection(URL url) throws IOException {
		HttpURLConnection conn;
		// URLConnection
		if(proxyHostName != null){
			Proxy proxy = new Proxy(proxyType, new InetSocketAddress(proxyHostName, proxyPort));
			conn = (HttpURLConnection) url.openConnection(proxy);
		}else{
			conn = (HttpURLConnection) url.openConnection();
		}

		conn.setReadTimeout(10 * 1000 * 60);
		if (getReadTimeout() >= 0) {
			conn.setReadTimeout(getReadTimeout());
		}
		// Enable writing to server ( to write request )
		conn.setDoOutput(true);
		// Enable reading from server ( to read response )
		conn.setDoInput(true);
		// Disable cache
		if(isCacheEnabled()){
			conn.setUseCaches(true);
			conn.setDefaultUseCaches(true);
		}else{
			conn.setUseCaches(false);
			conn.setDefaultUseCaches(false);
		}
		
		// Set method
		conn.setRequestMethod(HTTP_METHOD.POST.toString());
		conn.setInstanceFollowRedirects(isFollowRedirects());
		// Reset Authenticator
		try {
			// this block is written for suppressing a bug in sun implementation
			// In Sun Impl client doesn't authenticate user for each connection
			// , uses cached credentials instead
			@SuppressWarnings("unused")
			Class<?> cacheValClass = Class
					.forName("sun.net.www.protocol.http.AuthCacheValue");
			sun.net.www.protocol.http.AuthCacheValue
					.setAuthCache(new sun.net.www.protocol.http.AuthCacheImpl());

		} catch (Exception e) {

		}
		// Set headers
		conn.setRequestProperty(PROP_ACCEPT_ENCODING, getAcceptedEncodings());
		conn.setRequestProperty(PROP_USER_AGENT, getUserAgent());
		conn.setRequestProperty(PROP_ACCEPT, getAcceptedType());
		conn.setRequestProperty(PROP_ACCEPT_LANGUAGE, getAcceptedLanguage());
		conn.setRequestProperty(PROP_ACCEPT_CHARSET, getCharset());
		if (isKeepAlive()) {
			conn.setRequestProperty(PROP_KEEP_ALIVE, getKeepAliveTime()
					.toString());
			conn.setRequestProperty(PROP_CONNECTION,
					PROP_VAL_CONNECTION_KEEP_ALIVE);
		} else {
			conn.setRequestProperty(PROP_CONNECTION, PROP_VAL_CONNECTION_CLOSE);
		}
		conn.setRequestProperty(PROP_CONTENT_TYPE, getContentType());
		return conn;
	}

	/**
	 * same as sendData(method,null) 
	 */
	public HttpResponse sendData(HTTP_METHOD method) throws IOException{
		return sendData(method,null);
	}
	
	/**
	 * Sends the given String data to the URI with the given request method
	 * @param method Http request method 
	 * @param sendData the data to be used in the request
	 * @return Response of the server
	 * @throws IOException 
	 */
	public HttpResponse sendData(HTTP_METHOD method, String sendData)
			throws IOException {

		if(method == null){
			throw new IllegalArgumentException("HTTP Method is null!");
		}
		// Authentication
		if (this.uname != null) {
			TSAuthenticator.setCredentials(uri.toString(), uname, password);
		}
		if (this.proxyUname != null) {
			TSAuthenticator.setProxyCredentials(uri.toString(), proxyUname,
					proxyPassword);
		}
		try {
			String query = sendData != null && sendData.trim().length() > 0 ? uri
					.toString() + "?" + sendData
					: uri.toString();

			// Connection
			HttpURLConnection conn = (method == HTTP_METHOD.GET ? getConnection(new URL(
					query)) : getConnection(uri.toURL()));
			conn.setRequestMethod(method.toString());
			// Set additional props
			for (String propName : additionalRequestProperty.keySet()) {
				conn.setRequestProperty(propName,
						additionalRequestProperty.get(propName));
			}

			conn.setRequestProperty("Content-Length",
					String.valueOf(sendData != null ? sendData.length(): 0));
			// Connect
			conn.connect();
			HttpResponse res = null;
			if (sendData != null && method.equals(HTTP_METHOD.POST)) {
				// Write Request
				OutputStreamWriter wr = new OutputStreamWriter(
						conn.getOutputStream(), getCharset());
				wr.write(sendData);
				wr.flush();
				wr.close();
			}

			// Get the response

			// Get the content
			int responseCode = conn.getResponseCode();
			final int length = conn.getContentLength();
			String encoding = conn.getContentEncoding();

			if (responseCode == HttpURLConnection.HTTP_OK
					|| responseCode == 302) {
				InputStream stream = null;

				if (encoding == null)
					stream = conn.getInputStream();
				else if (encoding.trim().equalsIgnoreCase(ENCODING_GZIP))
					stream = new GZIPInputStream(conn.getInputStream());
				else if (encoding.trim().equalsIgnoreCase(ENCODING_DEFLATE))
					stream = new InflaterInputStream(conn.getInputStream(),
							new Inflater(true));
				else if (encoding.trim().equalsIgnoreCase(ENCODING_ZIP))
					stream = new ZipInputStream(conn.getInputStream());
				else
					stream = conn.getInputStream();

				if (stream != null) {
					int buflen = Math.max(getBufferSize(),
							Math.max(length, stream.available()));
					BufferedReader rd = new BufferedReader(
							new InputStreamReader(stream, getCharset()));
					// Read the request
					CharArrayWriter data = new CharArrayWriter();
					char buf[] = new char[buflen];
					int ret;
					while ((ret = rd.read(buf, 0, buflen)) != -1)
						data.write(buf, 0, ret);
					res = new HttpResponse(responseCode, data.toString(),conn.getResponseMessage(), getCookies(), conn.getHeaderFields());
					data.close();
					rd.close();
				}
			} else {
				InputStream stream = conn.getErrorStream();
				if (stream != null) {
					int buflen = Math.max(getBufferSize(),
							Math.max(length, stream.available()));
					BufferedReader rd = new BufferedReader(
							new InputStreamReader(stream, getCharset()));
					// Read the request
					CharArrayWriter data = new CharArrayWriter();
					char buf[] = new char[buflen];
					int ret;
					while ((ret = rd.read(buf, 0, buflen)) != -1)
						data.write(buf, 0, ret);
					res = new HttpResponse(true, responseCode, data.toString(), conn.getResponseMessage(),getCookies(), conn.getHeaderFields());
					data.close();
					rd.close();
				} else {
					res = new HttpResponse(true, responseCode, "", conn.getResponseMessage(),getCookies(), conn.getHeaderFields());
				}
			}
			conn.disconnect();
			return res;
		} catch (IOException ex) {
			throw ex;
		} finally {
			// Reset Authentication Info
			TSAuthenticator.setCredentials(uri.toString(), null, null);
			TSAuthenticator.setProxyCredentials(uri.toString(), null, null);
		}

	}

	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		cookieStore.removeAll();
	}

	@Override
	public String toString() {
		return getURI().toString();
	}
}
