package net.sf.corn.httpclient;

import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * A simple implementation of {@link java.net.CookieStore}, which stores
 * cookies for URI addresses in it
 * 
 * @author Serhat Dirik
 * 
 */
public class SimpleCookieStore implements CookieStore {

	private static final char PARAMETER_SPECIFIER = '?';

	private static Map<URI, List<HttpCookie>> cookieMap = new HashMap<URI, List<HttpCookie>>();

	public SimpleCookieStore() {
		removeAll();
	}

	/**
	 * Adds a cookie to be used with the given uri 
	 * 
	 * @param uri
	 *            uri target
	 * @param cookie
	 *            a cookie to be used for requests to given uri  
	 */
	public void add(URI uri, HttpCookie cookie) {
		URI _uri = null;
		if (uri.toString().indexOf(PARAMETER_SPECIFIER) > 0) {
			try {
				_uri = new URI(uri.toString().substring(0,
						uri.toString().indexOf(PARAMETER_SPECIFIER) - 1));
			} catch (URISyntaxException e) {
				_uri = uri;
			}
		} else {
			_uri = uri;
		}
		if (cookieMap.get(_uri) != null) {
			cookieMap.get(_uri).add(cookie);
		} else {
			List<HttpCookie> clist = new ArrayList<HttpCookie>();
			clist.add(cookie);
			cookieMap.put(_uri, clist);
		}
	}

	/**
	 * retrieves list of saved cookies for the given uri
	 * 
	 * @param uri
	 *            uri address
	 * @return list of previously set cookies
	 */
	public List<HttpCookie> get(URI uri) {
		URI _uri = null;
		if (uri.toString().indexOf(PARAMETER_SPECIFIER) > 0) {
			try {
				_uri = new URI(uri.toString().substring(0,
						uri.toString().indexOf(PARAMETER_SPECIFIER) - 1));
			} catch (URISyntaxException e) {
				_uri = uri;
			}
		} else {
			_uri = uri;
		}

		if (cookieMap.get(_uri) != null) {
			return cookieMap.get(_uri);
		} else {
			return new ArrayList<HttpCookie>();
		}
	}

	/**
	 * Retrieves whole cookie list in this store
	 */
	public List<HttpCookie> getCookies() {
		List<HttpCookie> retlist = new ArrayList<HttpCookie>();
		for (List<HttpCookie> clist : cookieMap.values()) {
			retlist.addAll(clist);
		}
		return retlist;
	}

	/**
	 * Retrieves list of addresses
	 */
	
	public List<URI> getURIs() {
		List<URI> retlist = new ArrayList<URI>();
		retlist.addAll(cookieMap.keySet());
		return retlist;
	}

	/**
	 * Removes a cookie that previously set
	 */
	
	public boolean remove(URI uri, HttpCookie cookie) {
		URI _uri = null;
		if (uri.toString().indexOf(PARAMETER_SPECIFIER) > 0) {
			try {
				_uri = new URI(uri.toString().substring(0,
						uri.toString().indexOf(PARAMETER_SPECIFIER) - 1));
			} catch (URISyntaxException e) {
				_uri = uri;
			}
		} else {
			_uri = uri;
		}

		if (cookieMap.get(_uri) != null) {
			return cookieMap.get(_uri).remove(cookie);
		} else {
			return true;
		}
	}

	public void addCookiesString(URI uri, String cookies) {
		URI _uri = null;
		if (uri.toString().indexOf(PARAMETER_SPECIFIER) > 0) {
			try {
				_uri = new URI(uri.toString().substring(0,
						uri.toString().indexOf(PARAMETER_SPECIFIER) - 1));
			} catch (URISyntaxException e) {
				_uri = uri;
			}
		} else {
			_uri = uri;
		}
		StringTokenizer st = new StringTokenizer(cookies, ";");
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			int inx = token.indexOf("=");
			if (inx > 0 && inx < token.length() - 1) {
				String name = token.substring(0, inx).trim();
				String value = token.substring(inx + 1, token.length()).trim();
				HttpCookie cookie = new HttpCookie(name, value);
				add(_uri, cookie);
			}
		}
	}

	/**
	 * gives all cookie values
	 * 
	 * @return cookies as one string
	 */
	public String getCookiesString(URI uri) {
		URI _uri = null;
		if (uri.toString().indexOf(PARAMETER_SPECIFIER) > 0) {
			try {
				_uri = new URI(uri.toString().substring(0,
						uri.toString().indexOf(PARAMETER_SPECIFIER) - 1));
			} catch (URISyntaxException e) {
				_uri = uri;
			}
		} else {
			_uri = uri;
		}

		List<HttpCookie> cookies = get(_uri);
		StringBuilder sb = new StringBuilder();
		boolean firstOne = true;
		for (HttpCookie cookie : cookies) {
			String cname = cookie.getName();
			if (!firstOne)
				sb.append(";");
			else
				firstOne = false;
			String cval = cookie.getValue();
			sb.append(cname);
			sb.append("=");
			sb.append(cval);
		}
		return sb.toString();
	}

	/**
	 * Removes all cookies in the store
	 */
	
	public boolean removeAll() {
		cookieMap.clear();
		return true;
	}

	
	protected void finalize() throws Throwable {
		super.finalize();
		removeAll();
	}

	
	public String toString() {
		return cookieMap.toString() ;
	}
}
