package net.sf.corn.httpclient;

import java.net.Authenticator;
import java.net.PasswordAuthentication;
import java.net.URI;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * This class is a thread safe implementation of {@link java.net.Authenticator}
 * class. It keeps a thread safe address and credentials pairs map to use for http
 * client requests. You can set credentials for a specific address or a regular expression 
 * that will be used to matching at runtime. The credentials that saved for specific addresses
 * always has the highest priority to be used with. Authenticator also permits to save
 * default credentials to use with any requested URL. 
 * 
 * @author Serhat Dirik
 * @see java.net.Authenticator
 */
public class TSAuthenticator extends Authenticator {

	private static final char PARAMETER_SPECIFIER = '?';

	/*
	 * Username password credential
	 */
	private static class UPCredential {
		private transient String uname = null;
		private transient String password = null;

		public UPCredential(String uname, String password) {
			super();
			this.uname = uname;
			this.password = password;
		}

		public String getUname() {
			return uname;
		}

		public String getPassword() {
			return password;
		}
	}

	/**
	 * URL Mapped security credentials
	 * 
	 * @author <a href="mailto:serhat.dirik@google.com">Serhat Dirik</a>
	 * 
	 */
	private static class TLCredentialsMap extends
			ThreadLocal<Map<String, UPCredential>> {
		@Override
		protected Map<String, UPCredential> initialValue() {
			return null;
		}

		@Override
		public Map<String, UPCredential> get() {
			return super.get();
		}

		@Override
		public void set(Map<String, UPCredential> value) {
			super.set(value);
		}
	}

	private static class TLCredential extends ThreadLocal<UPCredential> {
		@Override
		protected UPCredential initialValue() {
			return null;
		}

		@Override
		public UPCredential get() {
			return super.get();
		}

		@Override
		public void set(UPCredential value) {
			super.set(value);
		}
	}

	private static TLCredentialsMap tlUrlCredentialsMap = new TLCredentialsMap();
	private static TLCredentialsMap tlUrlProxyCredentialsMap = new TLCredentialsMap();

	private static TLCredentialsMap tlPatternCredentialsMap = new TLCredentialsMap();
	private static TLCredentialsMap tlPatternProxyCredentialsMap = new TLCredentialsMap();

	private static TLCredential tlDefaultCredential = new TLCredential();
	private static TLCredential tlDefaultProxyCredential = new TLCredential();

	private TSAuthenticator() {
	}

	/**
	 * Sets default security credentials to pass proxy server
	 * 
	 * @param uname
	 * @param password
	 */
	public static void setDefaultProxyCredential(String uname, String password) {
		tlDefaultProxyCredential.set(new UPCredential(uname, password));
	}

	/**
	 * Sets default security credentials
	 * 
	 * @param uname
	 * @param password
	 */
	public static void setDefaultCredential(String uname, String password) {
		tlDefaultCredential.set(new UPCredential(uname, password));
	}

	/**
	 * Sets security credentials to pass proxy server for the given url target
	 * 
	 * @param url
	 *            URL string
	 * @param uname
	 *            user name
	 * @param password
	 *            password
	 */
	public static void setProxyCredentials(String url, String uname,
			String password) {
		if (url == null) {
			throw new IllegalArgumentException("Null url value in");
		}
		String _url = null;
		if (url.indexOf(PARAMETER_SPECIFIER) > 0) {
			_url = url.substring(0, url.indexOf(PARAMETER_SPECIFIER) - 1);
		} else {
			_url = url;
		}
		putInMap(tlUrlProxyCredentialsMap, _url, uname, password);
	}

	/**
	 * Sets security credentials to pass proxy server for the given url target
	 * 
	 * @param url
	 *            URL string
	 * @param uname
	 *            user name
	 * @param password
	 *            password
	 */
	public static void setProxyCredentials(URL url, String uname,
			String password) {
		if (url == null) {
			throw new IllegalArgumentException("Null url value in");
		}
		setProxyCredentials(url.toString(), uname, password);
	}

	/**
	 * Sets security credentials to pass proxy server for the given URI target
	 * 
	 * @param uri
	 *            URI address
	 * @param uname
	 *            user name
	 * @param password
	 *            password
	 */
	public static void setProxyCredentials(URI uri, String uname,
			String password) {
		if (uri == null) {
			throw new IllegalArgumentException("Null uri value in");
		}
		setProxyCredentials(uri.toString(), uname, password);
	}

	/**
	 * Sets proxy credentials for any url that matches with the given regular
	 * expression pattern
	 * 
	 * @param regexPattern
	 *            Regular expression text
	 * @param uname
	 *            user name
	 * @param password
	 *            password for the given user
	 */
	public static void setProxyCredentialsForPattern(String regexPattern,
			String uname, String password) {
		if (regexPattern == null) {
			throw new IllegalArgumentException("Null regexPattern value in");
		}
		putInMap(tlPatternProxyCredentialsMap, regexPattern, uname, password);
	}

	/**
	 * Sets security credentials for the url address
	 * 
	 * @param url   url address
	 * @param uname user name 
	 * @param password password for the given user
	 */
	public static void setCredentials(String url, String uname, String password) {
		if (url == null) {
			throw new IllegalArgumentException("Null url value in");
		}
		String _url = null;
		if (url.indexOf(PARAMETER_SPECIFIER) > 0) {
			_url = url.substring(0, url.indexOf(PARAMETER_SPECIFIER) - 1);
		} else {
			_url = url;
		}
		putInMap(tlUrlCredentialsMap, _url, uname, password);
	}

	/**
	 * Sets security credentials for the url 
	 * 
	 * @param url 
	 * @param uname
	 * @param password
	 */
	public static void setCredentials(URL url, String uname, String password) {
		if (url == null) {
			throw new IllegalArgumentException("Null url value in");
		}
		setCredentials(url.toString(), uname, password);
	}
	
	/**
	 * Sets security credentials for the given URI 
	 * 
	 * @param uri
	 * @param uname
	 * @param password
	 */
	public static void setCredentials(URI uri, String uname, String password) {
		if (uri == null) {
			throw new IllegalArgumentException("Null URI value in");
		}
		setCredentials(uri.toString(), uname, password);
	}
	
	
	
	/**
	 * Sets credentials for any url that matches with the given regular
	 * expression pattern
	 * 
	 * @param regexPattern
	 *            Regular expression text
	 * @param uname
	 *            user name
	 * @param password
	 *            password for the given user
	 */
	public static void setCredentialsForPattern(String regex, String uname,
			String password) {
		if (regex == null) {
			throw new IllegalArgumentException("Null regex value in");
		}
		putInMap(tlPatternCredentialsMap, regex, uname, password);
	}

	/**
	 * Sets this class as the default authenticator in the current JVM
	 */
	@SuppressWarnings({ "restriction", "unused" })
	public static void setAsDefault() {
		Authenticator.setDefault(new TSAuthenticator());
		try {
			// this block is written for suppressing a bug in sun implementation
			// In Sun Impl client doesn't authenticate user for each connection
			// , uses cached credentials instead
			Class<?> cacheValClass = Class
					.forName("sun.net.www.protocol.http.AuthCacheValue");
			sun.net.www.protocol.http.AuthCacheValue
					.setAuthCache(new sun.net.www.protocol.http.AuthCacheImpl());
		} catch (Exception e) {
		}
	}

	private static void putInMap(TLCredentialsMap map, String url,
			String uname, String password) {
		if (uname == null) {
			if (map.get() != null) {
				map.get().remove(url);
				if (map.get().size() == 0)
					map.set(null);
			}
			return;
		}
		UPCredential credential = new UPCredential(uname, password);
		if (map.get() == null) {
			map.set(new HashMap<String, TSAuthenticator.UPCredential>());
		}
		map.get().put(url, credential);
	}

	private UPCredential findProperCredential(String url,
			TLCredentialsMap urlMap, TLCredentialsMap patternMap,
			TLCredential defaultCredential) {
		UPCredential credential = null;
		credential = urlMap.get() != null ? urlMap.get().get(url) : null;

		if (credential == null) {
			Map<String, UPCredential> map = patternMap.get();
			Set<String> keyset = map != null ? map.keySet() : null;
			if (keyset != null) {
				for (String pattern : keyset) {
					if (url.matches(pattern)) {
						credential = map.get(pattern);
						break;
					}
				}
			}
		}
		if (credential == null) {
			credential = defaultCredential.get();
		}
		return credential;
	}

	protected PasswordAuthentication getPasswordAuthentication() {
		String _url = getRequestingURL().toExternalForm();
		if (_url.indexOf(PARAMETER_SPECIFIER) > 0) {
			_url = _url.substring(0, _url.indexOf(PARAMETER_SPECIFIER));
		} else {
			_url = _url.trim();
		}

		UPCredential credential = null;
		TLCredentialsMap urlMap = null;
		TLCredentialsMap patternMap = null;
		TLCredential defaultCredential = null;

		if (getRequestorType().equals(RequestorType.PROXY)) {
			urlMap = tlUrlProxyCredentialsMap;
			patternMap = tlPatternProxyCredentialsMap;
			defaultCredential = tlDefaultProxyCredential;
		} else {
			urlMap = tlUrlCredentialsMap;
			patternMap = tlPatternCredentialsMap;
			defaultCredential = tlDefaultCredential;
		}
		credential = findProperCredential(_url, urlMap, patternMap,
				defaultCredential);

		if (credential == null && _url.endsWith("/")) {
			credential = findProperCredential(
					_url.substring(0, _url.length() - 1), urlMap, patternMap,
					defaultCredential);
		}

		if (credential != null) {
			return new PasswordAuthentication(credential.getUname(), credential
					.getPassword().toCharArray());
		} else {
			return null;
		}
	}

	/**
	 * Resets all credentials that previously set
	 */
	public static void reset() {
		tlDefaultCredential.set(null);
		tlDefaultProxyCredential.set(null);
		tlPatternCredentialsMap.set(null);
		tlPatternProxyCredentialsMap.set(null);
		tlUrlCredentialsMap.set(null);
		tlUrlProxyCredentialsMap.set(null);
	}
}
